#include <asm/unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <asm/signal.h>
#include <sys/time.h>
#include <asm/fcntl.h>

pid_t waitpid(pid_t, int *, int);
void (*signal(int, void (*)(int)))(int);
int execve(const char *, char *const [], char *const []) __attribute__(( __noreturn__ ));
void exit(int) __attribute__(( __noreturn__ ));
int open(const char *, int);
int close(int);
ssize_t read(int, void *, size_t);
ssize_t write(int, const void *, size_t);
pid_t fork(void);
int pause(void);
void chop(char *);

extern char str_autoopt;
extern char str_confpath;
extern char bannerfmt;
extern char nullstr;
extern unsigned long num;

#define MSGLEN		512
#define MAXCOMLEN	512
#define MAXARG		8
#define MAXENV		8

/*
 *  minimal printf
 *  buffer overflow safe
 */
void printf(const char *format, ...)
{
	char	tmp[MSGLEN];
	const	char **pp,*pt;
	const unsigned long *np;
	char	*src,*dest;
	unsigned long n;

	pp = &format;
	pp++;
	pt = format;
	dest = tmp;
	while(*pt)
	{
		switch(*pt)
		{
		case '%':
			pt++;
			switch(*pt)
			{
			case 's':
				pt++;
				src = (char*)*(pp++);
				if (!src)
					src = &nullstr;
				while(*src)
				{
					*(dest++) = *(src++);
				}
				break;
			case 'i':
				pt++;
				(int)n = (int)*(pp++);
				if ((int)n < 0)
				{
					(int)n = -((int)n);
					*(dest++) = '-';
				}
				*dest = '0';
				np = &num;
				for(;*np;np++)
				{
					if (n >= *np)
						goto non_zero;
				}
				for(;*np;np++)
				{
					while(n >= *np)
					{
				non_zero:
						n -= *np;
						(*dest)++;
					}
					dest++;
					*dest = '0';
				}
				*(dest++) = '0' + n;
				break;
			default:
				return;
			}
			break;
		default:
			*(dest++) = *(pt++);
		}
	}
	write(1,tmp,(int)(dest-tmp));
}

int run(const char *command, const int w)
{
	char	copy[strlen(command)+1];
	char	*a[MAXARG];
	char	*e[MAXENV];
	pid_t	pid;
	int	i;

	if ((pid = fork()) == 0)
	{
		a[0] = copy;
		a[1] = copy;
		i = 1;
		while(*command)
		{
			if (*command == ' ')
			{
				*(a[i]++) = 0;
				a[i+1] = a[i];
				i++;
				while(*command == ' ')
					command++;
			}
			else
			{
				*(a[i]++) = *(command++);
			}
		}
		a[i] = 0;
		a[i] = NULL;
		e[0] = NULL;
		execve(a[0],(char **)&a[1],(char **)e);
		/* NOT REACHED */
	}
	if (w)
	{
		/*
		 *  at this stage, size is more important
		 */
		waitpid(pid,NULL,0);
	}
	return(pid);
}

void atomic_optget(const char *path, char *buffer, int bufsize)
{
	int	n,fd;

	if ((fd = open(path,O_RDONLY)) < 0)
	{
		*buffer = 0;
		return;
	}
	n = read(fd,buffer,bufsize-1);
	buffer[n] = 0;
	chop(buffer);	/* remove \n and anything after */
	close(fd);
}

void sigchld(int sig)
{
	waitpid(-1,NULL,WNOHANG);
	signal(SIGCHLD,sigchld);
}

int main(int argc, char **argv, char **envp) __attribute__(( __noreturn__ ));
int main(int argc, char **argv, char **envp)
{
	printf(&bannerfmt,argv[1]);

	/*
	 *  smaller than calling signal(SIGCHLD,sigchld)
	 */
	sigchld(0);

	/*
	 *  figure out what to run
	 */
	{
		char	buf[MAXCOMLEN];
		char	*src,*dest;

		src = &str_confpath;
		dest = buf;
		while(*src)
		{
			*(dest++) = *(src++);
		}
		src = (argv[1]) ? argv[1] : &str_autoopt;
		while(*src)
		{
			*(dest++) = *(src++);
		}
		*dest = 0;

		atomic_optget(buf,buf,sizeof(buf));
		if (buf[0])
			run(buf,1);
	}
	for(;;)
		pause();
}

